/**********************************************************************/
/*                                                                    */
/*        hist.c -- a tcsh like input function                        */
/*                                                                    */
/*                by K.sugiura   Ver 1.0  for GEDO of LINUX           */
/*                               freezed at 1 Nov. 1996               */
/*                                                                    */
/*        export functions                                            */
/*                                                                    */
/*           voiid  ctrl_c(int signo);                                */
/*           void  init_raw_tty(void);                                */
/*           void reset_raw_tty(void);                                */
/*           void  end_raw_tty(void);                                 */
/*           void  init_ring(void);                                   */
/*           void  set_ring( char *str, int len );                    */
/*           void  get_str(char  *str,int  max,int  pos);             */
/*                                                                    */
/*        export variables                                            */
/*                        NONE                                        */
/*        B0 1.16,1997   History adjustment get_str()                 */
/*        rev D2:  1999.12.11: SUSPEND(Ctrl+Z) allowance              */
/*        rev D2:  2000.2.8: change signal to sigaction.              */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <memory.h>
#include <termio.h>
#include <curses.h>
#include <signal.h>
#include <setjmp.h>
#include <sys/ioctl.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/types.h>
#include <errno.h>
#include "gedo.h"
#include "prot.h"
#include "error.h"

#ifdef TERMCAP
#include <termcap.h>
#endif

#ifdef TERMINFO
#include <term.h>
#endif

/* KEYCODE assign */
#define UP	0x10
#define DOWN	0x0e

#define	BS	0x08
#define TAB	0x09
#define RET	0x0a
#define RIGHT	0x06
#define LEFT	0x02
#define INS	0x12
#define DEL	0x04
#define TOP	0x01
#define TAIL	0x05
#define DLINE   0x0b


#define INSERT	0
#define OVERW	1
#define CHANGE	1
#define NOCHANGE  0

#define RINGSIZE 8192	/* size of RING Buffer */
#define BUFFSIZE RINGSIZE	/* size of tmp buffer */
#define TERMBUFF 1024	/* termcap conversion work area */
#define LINESIZE 1024	/* replace_dash work area */
#define TABSIZE   4	/* TAB(0x09) expandion to SPACE*TABSIZE */


#ifdef TERMCAP
#define c_next()  tputs(cursor_right,1,outc)
#define c_back()  tputs(cursor_left,1,outc)
#define c_up()    tputs(cursor_up,1,outc)
#define c_down()  tputs(cursor_down,1,outc)
#define c_clear()   tputs(clr_eos,1,outc)

static char *cursor_right, *cursor_left, *cursor_up, *cursor_down;
static char *clr_eos;
static int columns, lines;
#endif


#ifdef TERMINFO
#define c_next()  putp(cursor_right)
#define c_back()  putp(cursor_left)
#define c_up()    putp(cursor_up)
#define c_down()  putp(cursor_down)
#define c_clear()   putp(clr_eos)
#endif


#define r_inc(i)	((i)>=RINGSIZE-1)?((i)=0):((i)++)
#define r_dec(i)	((i)<=0)?((i)=RINGSIZE-1):((i)--)
#define rp_inc(p)	((p)>=ring+RINGSIZE-1)?((p)=ring):((p)++)

static char ring[RINGSIZE];	/* RING buffer(Keybord History) */
static int sp;			/* ring buffer adding pointer */
static int at;			/* ring buffer now pointer */
static int next;		/* ring buffer next pointer */
static int prev;		/* ring buffer previous pointer */
static int first;		/* ring buffer first flag */


static struct termio Org;	/* termio struct */
static int  Raw_flag = 0;	/* tty mode */
static void (*OldTSTP)( int signo ); /* old hander to SIGTSTP */
static void (*OldCONT)( int signo );  /* old hander to SIGCONT */
void TSTP_handler(int signo);     /* prototype to handlers */
void CONT_handler(int signo);


/**********************************************************************/
/*                                                                    */
/*             function mysignal()                                    */
/*                                                                    */
/*             int sig: signal to catch                               */
/*             sighandler_t hand: signal handler to set               */
/*                                                                    */
/*             return  NONE                                           */
/*                                                                    */
/*             function: substitute of signal, by POSIX.              */
/*             D2: create new                                         */
/**********************************************************************/
/* At TurboLinux(libc 2.0.7), action of signal is different from before.
 If GUI response is here, Calling getchar(3) in input_char() is NOT break
 by EINTR at new version.
   So I want system calls to break by signal. POSIX version to set signal
 handler is sigaction(2), this syscall is defined always BREAK by signal.
   This syscall is what I need. (But can be NOT break in Linux alternativly,
 set SA_RESTART to sa.sa_flgs, here syscalls may not be breaked by signal.
 */   
void mysignal( int sig, __sighandler_t hand )
{
     struct sigaction sa;
     sa.sa_handler = hand;
     sigemptyset( &sa.sa_mask );
     sa.sa_flags = 0;
     sigaction( sig, &sa, NULL );
}

/**********************************************************************/
/*                                                                    */
/*             function outc()                                        */
/*                                                                    */
/*             int c: character code                                  */
/*                                                                    */
/*             return  int: character code                            */
/*                                                                    */
/*             function: wrapper for TERMCAP                          */
/*                                                                    */
/**********************************************************************/
int outc( int c )
{
	return( putchar( c ) );
}

/**********************************************************************/
/*                                                                    */
/*             function set_org_termio()                              */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: reset to COOKED tty                          */
/*                                                                    */
/**********************************************************************/
void set_org_termio( void )
{
	ioctl( 0, TCSETA, &Org );
}


/**********************************************************************/
/*                                                                    */
/*             function init_raw_tty()                                */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: GET curses function and CHANGE to RAW mode   */
/*                                                                    */
/**********************************************************************/
void init_raw_tty()
{
	char *term;
	struct termio new;
	char entbuf[TERMBUFF];
	static char sgbuf[TERMBUFF];
	char *bufptr;
#ifdef TERMINFO
	int status;
#endif

	if( (term = getenv("TERM")) == NULL ){
		fatal( F_NOTERM );
	}
#ifdef TERMCAP
	if( tgetent( entbuf, term ) != 1 ){
		fatal( F_TERM );
	}
	lines = tgetnum( "li" );
	columns = tgetnum( "co" );
	bufptr = sgbuf;
	cursor_right = tgetstr( "cl", &bufptr );

	if( (cursor_right = tgetstr( "nd", &bufptr ))  == NULL ){
		fatal( F_TERM );
	}
	if( (cursor_left = tgetstr( "le", &bufptr )) == NULL ){
		fatal( F_TERM );
	}
	if( (cursor_up = tgetstr( "up", &bufptr )) == NULL ){
		fatal( F_TERM );
	}
	if( (cursor_down = tgetstr( "do", &bufptr )) == NULL ){
		fatal( F_TERM );
	}
	if( (clr_eos = tgetstr( "cd", &bufptr )) == NULL ){
		fatal( F_TERM );
	}
#endif
#ifdef TERMINFO
	setupterm( term, 1, &status );
	if( status != 1 ){
		fatal( F_TERM );
	}

	if( cursor_right  == NULL ){
		fatal( F_TERM );
	}
	if( cursor_left == NULL ){
		fatal( F_TERM );
	}
	if( cursor_up  == NULL ){
		fatal( F_TERM );
	}
	if( cursor_down == NULL ){
		fatal( F_TERM );
	}
	if( clr_eos == NULL ){
		fatal( F_TERM );
	}
#endif

	if( ioctl( 0, TCGETA, &Org ) == EOF ){
		fatal( F_IOCTL );
	}
	memcpy( &new, &Org, sizeof(struct termio) );
	new.c_lflag &= ~ECHO;
	new.c_lflag &= ~ICANON;
	new.c_cc[VMIN] = 1;
	new.c_cc[VTIME] = 0;
	if( ioctl( 0, TCSETA, &new ) == EOF ){
		fatal( F_IOCTL );
	}
/* D0 correctly handle after Ctrl+Z */
	OldTSTP = signal( SIGTSTP, TSTP_handler );
	OldCONT = signal( SIGCONT, CONT_handler );
	Raw_flag = 1;
}

void TSTP_handler( int signo )
{
     end_raw_tty();
     signal( SIGCONT, CONT_handler );
     if( OldTSTP == SIG_DFL ){
	  signal( SIGTSTP, SIG_DFL );
	  kill( getpid(), SIGTSTP );
     } else {
	  signal( SIGTSTP, TSTP_handler );
     }
}

void CONT_handler( int signo )
{
     reset_raw_tty();
     signal( SIGTSTP, TSTP_handler );
     if( OldCONT == SIG_DFL ){
	  signal( SIGCONT, SIG_DFL );
	  kill( getpid(), SIGCONT );
     } else {
	  signal( SIGCONT, CONT_handler );
     }
}

/**********************************************************************/
/*                                                                    */
/*             function reset_raw_tty()                               */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Resotre RAW mode                             */
/*                                                                    */
/**********************************************************************/
void reset_raw_tty()
{
	struct termio new;

	if( Raw_flag == 0 ){
		memcpy( &new, &Org, sizeof(struct termio) );
		new.c_lflag &= ~ECHO;
		new.c_lflag &= ~ICANON;
		new.c_cc[VMIN] = 1;
		new.c_cc[VTIME] = 0;
		if( ioctl( 0, TCSETA, &new ) == EOF ){
			fatal( F_IOCTL );
		}
		Raw_flag = 1;
	}
}

/**********************************************************************/
/*                                                                    */
/*             function end_raw_tty()                                 */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Change to COOKED tty                         */
/*                                                                    */
/**********************************************************************/
void end_raw_tty()
{
	if( Raw_flag ){
		if( ioctl( 0, TCSETA, &Org ) == EOF ){
			fatal( F_IOCTL );
		}
		Raw_flag = 0;
	}
}

/**********************************************************************/
/*                                                                    */
/*             function init_ring()                                   */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Intialize Ring Buffer.                       */
/*                                                                    */
/**********************************************************************/
void init_ring()
{
	prev = next = sp = at = 0;
	first = 1;
	memset( ring, 0, sizeof(ring) );
}

/**********************************************************************/
/*                                                                    */
/*             function get_ring()                                    */
/*                                                                    */
/*             int flag: UP or DOWN                                   */
/*             char *str: buffer                                      */
/*             int max: buffer maximum                                */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: get form History(UP or DOWN).                */
/*                                                                    */
/**********************************************************************/
static void get_ring( int flag, char *str, int max )
{
	char *p;

	if( first == 1 ){
		*str = '\0';
		return;
	}
	if( flag == UP ){
		for( p = &ring[prev]; (*str++ = *p); rp_inc(p) );
		next = at;
		at = prev;
		r_dec(prev);
		for( ; ring[prev] == 0; r_dec(prev) );
		for( ; ring[prev] != 0; r_dec(prev) );
		r_inc(prev);
	} else if( flag == DOWN ){
		for( p = &ring[next]; (*str++ = *p); rp_inc(p) );
		prev = at;
		at = next;
		for( ; ring[next] != 0; r_inc(next) );
		for( ; ring[next] == 0; r_inc(next) );
	}
	return;
}

/**********************************************************************/
/*                                                                    */
/*             function set_ring()                                    */
/*                                                                    */
/*             char *str: buffer                                      */
/*             int len:   length of str                               */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Set str to History.                          */
/*                                                                    */
/**********************************************************************/
void set_ring( char *str, int len )
{
	int i;

	for( i = len-1; i >= 0; i-- ){
		if( str[i] == ' ' ){
			str[i] = 0;
		} else {
			break;
		}
	}
	if( i < 0 ){
		return;
	}
	first = 0;
	prev = sp;
	for( ; *str; str++ ){
		ring[sp] = *str;
		r_inc(sp);
	}
	if( ring[sp] != 0 ){
		ring[sp] = 0;
		r_inc(sp);
		for( i = sp; ring[i]; r_inc(i) ){
			ring[i] = 0;
		}
		for( ; ring[i] == 0; r_inc(i) );
		next = i;
	} else {
		r_inc(sp);
	}
	at = sp;
}

/**********************************************************************/
/*                                                                    */
/*             function replace_dash()                                */
/*                                                                    */
/*             char *str: <IN/OUT> input string                       */
/*                                                                    */
/*             return  int: ==1 found replace  ==0 cannot find replace*/
/*                                                                    */
/*             function: find from History !! or !<a>, and substitute.*/
/*                                                                    */
/**********************************************************************/
static int replace_dash( char *str )
{
	int i;
	static int prevcom = -1;
	int prevtmp, attmp, nexttmp;
	char work[LINESIZE];
	char work2[LINESIZE];
	int ret;

	ret = 0;
	for( i = 0; str[i]; i++ ){
		if( str[i] == '!' ){
			if( str[i+1] == '!' ){
				if( prevcom == -1 ){
					break;
				}
				prevtmp = prev;
				attmp = at;
				nexttmp = next;
				prev = prevcom;
				get_ring( UP, work, sizeof(work) );
			        prev = prevtmp;
				at = attmp;
				next = nexttmp;
				strcpy( work2, &str[i+2] );
				strcpy( &str[i], work );
				strcat( &str[i], work2 );
				display( M_OMNI, str );
				ret = 1;
				break;
			} else {
				if( i != 0 ){
					break;
				}
				prevtmp = prev;
				attmp = at;
				nexttmp = next;
				prev = prevcom;
				get_ring( UP, work, sizeof(work) );
				if( strncmp( &str[1], work, strlen( &str[1] ) ) == 0 ){
					strcpy( str, work );
					display( M_OMNI, str );
					ret = 1;
					prev = prevtmp;
					at = attmp;
					next = nexttmp;
					break;
				}
				while( prevtmp != prev ){
					get_ring( UP, work, sizeof(work) );
					if( strncmp( &str[1], work, strlen( &str[1] ) ) == 0 ){
						strcpy( str, work );
						display( M_OMNI, str );
						ret = 1;
						break;
					}
				}
				prev = prevtmp;
				at = attmp;
				next = nexttmp;
			}
		}
	}
	prevcom = at;
	return( ret );
}



/**********************************************************************/
/*                                                                    */
/*             function specialkey()                                  */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return  int: Local Key Code                            */
/*                                                                    */
/*             function: Special Key Scan Code to Local Key Code.     */
/*                                                                    */
/**********************************************************************/

/* this function is LINUX local. Change codes for your machine! */
static int specialkey( void )
{
	int ret;

	ret = getchar();
	switch( ret ){
	      case 0x32:
		getchar();
		return( INS );
		break;
	      case 0x33:
		getchar();
		return( DEL );
		break;
	      case 0x41:
		return( UP );
		break;
	      case 0x42:
		return( DOWN );
		break;
	      case 0x43:
		return( RIGHT );
		break;
	      case 0x44:
		return( LEFT );
		break;
	      default:
		return( -1 );
		break;
	}
}

/* for Event Driven Insert routine used IPC msg */
#define GUIBUFFSIZE 256		/* Buffer for GUI receive */
static int Sem = 0;		/* Like Semaphone of GUI receive */
/* D2: new signals: not need*/
/*  static jmp_buf keybuf;		GUI interrupt */
struct smes {
	long mtype;
	char str[GUIBUFFSIZE];
};
struct smes Mes;		/* IPC msg buff */

extern int Gui;			/* GUI msg ID */
char GuiBuff[GUIBUFFSIZE];	/* Buffer for GUI */
char *Grp = GuiBuff;		/* LIFO read pointer */
char *Gwp = GuiBuff;		/* LIFO write pointer */

/**********************************************************************/
/*                                                                    */
/*             function catch()                                       */
/*                                                                    */
/*             int signo: interrupt signal no                         */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: signal handler for SIGUSR2(GUI interrupt)    */
/*                                                                    */
/**********************************************************************/
void catch( int signo )
{
	int len;

/* D2: new signal: return and break getchar() */
/*     NOT need to reset myself at sighandler, if sigaction(2) used. */
/* del	signal( SIGUSR1, SIG_IGN ); */
	if( Gui ){
		while( 	msgrcv( MesId, (struct msgbuf *)&Mes, GUIBUFFSIZE, Gui, IPC_NOWAIT ) != -1 ){
			len = strlen( Mes.str );
			if( Gwp + len < GuiBuff + sizeof(GuiBuff) ){
				strcpy( Gwp, Mes.str );
				Sem += len;
				Gwp += len;
			}
			memset( Mes.str, 0, sizeof(Mes.str) );
		}
	}
	if( check_value( GLOBAL, "#guikeycase" ) ){
	     getchar(); /* read & discard at interrupt 0x0a! */
	}
/* D2: new signal: return and break getchar() */
/*    This is NO GOOD, not return sighandlers but longjmp */
/* delete	longjmp( keybuf, 1 ); */
}

/**********************************************************************/
/*                                                                    */
/*             function check_guibuff()                               */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return  int: Now number of chars in GUIbuffer          */
/*                                                                    */
/*             function: check the GUI buffer                         */
/*                                                                    */
/**********************************************************************/
static int check_guibuff( void )
{
	int len;

	if( Gui == 0 ){
		return( 0 );
	}
	while( 	msgrcv( MesId, (struct msgbuf *)&Mes, GUIBUFFSIZE, Gui, IPC_NOWAIT ) != -1 ){
		len = strlen( Mes.str );
		if( Gwp + len < GuiBuff + sizeof(GuiBuff) ){
			strcpy( Gwp, Mes.str );
			Sem += len;
			Gwp += len;
		}
		memset( Mes.str, 0, sizeof(Mes.str) );
	}
	return( Sem );
}

/**********************************************************************/
/*                                                                    */
/*             function get_guichar()                                 */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return  int: character                                 */
/*                                                                    */
/*             function: get character from GUI buffer.               */
/*                                                                    */
/**********************************************************************/
static int get_guichar( void )
{
	int ret;

	ret = *Grp++;
	if( Grp >= GuiBuff + sizeof(GuiBuff) ){
		Grp = 0;
	}
	if( Grp == Gwp ){
		Gwp = Grp = GuiBuff;
		memset( GuiBuff, 0, sizeof(GuiBuff) );
	}
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function input_char()                                  */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return  int: character(Local Key Code)                 */
/*                                                                    */
/*             function: read a char from stdin or GUI.               */
/*                                                                    */
/**********************************************************************/
static int input_char()
{
	int ret;

	while( 1 ){
		if( Sem == 0 ){
			if( check_guibuff() ){
				ret = get_guichar();
				Sem--;
			} else {
/* D2 changed */
/* At TurboLinux(libc 2.0.7), action of signal is different from before.
 If GUI response is here, Calling getchar(3)  is NOT break by EINTR at 
 new version.
   So I want system calls to break by signal. POSIX version to set signal
 handler is sigaction(2), this syscall is defined always BREAK by signal.
   This syscall is what I need. (But can be NOT break in Linux alternativly,
 set SA_RESTART to sa.sa_flgs, here syscalls may not be breaked by signal.
   And change SIGUSR1 to NOTUSED SIGUSR2, this is safty and check check
 errno if INTERRUPT is here or not.*/   
/*				signal( SIGUSR1, catch );*/
/*				ret = getchar();*/
/*				signal( SIGUSR1, SIG_IGN );*/
/* replaced as follow... */
				mysignal( SIGUSR2, catch );
				ret = getchar();
				mysignal( SIGUSR2, SIG_IGN );
				if( errno == EINTR ){
				     continue;
				}
/* D2 end */
			}
		} else {
			ret = get_guichar();
			Sem--;
		}
		if( ret == 0x1b ){
			ret = getchar();
			if( ret == 0x5b ){
				ret = specialkey();
				if( ret != -1 ){
					return( ret );
				}
			}
		} else if( ret < 0x20 ){
			return( ret );
		} else if( ret == 0x7f ){
			return( BS );
		} else {
			return( ret );
		}
	}
}


/**********************************************************************/
/*                                                                    */
/*             function move_cursor()                                 */
/*                                                                    */
/*             int from: position jump from                           */
/*             int toat: position jump to                             */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: move cursor from <from> to <toat>.           */
/*                      (byte from top)                               */
/**********************************************************************/
static void move_cursor( int from, int toat )
{
	int fx, fy;
	int tx, ty;
	int x, y;

	if( from % columns == 0 ){
		from--;
	} 
	fx = from % columns;
	fy = from / columns;
	tx = toat % columns;
	ty = toat / columns;
	if( (y = fy - ty) > 0 ){
		while( y-- > 0 ){
			c_up();
		}
	} else if( y < 0 ){
		y = y * (-1);
		while( y-- > 0 ){
			c_down();
		}
	}
	if( (x = fx - tx) > 0 ){
		while( x-- > 0 ){
			c_back();
		}
	} else if( x < 0 ){
		x = x * (-1);
		while( x-- > 0 ){
			c_next();
		}
	}
}

static int flush_after( char *s, int from, int toat, int top, int flag )
{
	int j;

	for( j = from; j < toat; j++ ){
		if( flag == 0 ){
			if( ((top + j) % columns) == 0){
				flag = 1;
				putchar( '\n' );
			}
		}
		putchar( s[j] );
	}
	if( flag == 2 ){
		putchar( ' ' );
	}
	return( flag );
}


/**********************************************************************/
/*                                                                    */
/*             function get_str()                                     */
/*                                                                    */
/*             char *str: buffer                                      */
/*             int max:   maximum of buffer                           */
/*             int pos:   offset of 1st char(prompt length)           */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Read the input string.                       */
/*                                                                    */
/**********************************************************************/
void get_str( char *str, int max, int pos )
{
	int i, j;
	int c;
	int len, tabn;
	char buff[BUFFSIZE];
	static int mode = INSERT;
	int change_flag = NOCHANGE;
	int prevtmp, attmp, nexttmp;
	int breaked;

	i = len = breaked = 0;
	memset( str, 0, max );
	prevtmp = prev;
	attmp = at;
	nexttmp = next;
	while( 1 ){
		if( i >= max ){
			i = max -1;
			str[i] = '\0';
			move_cursor( max+pos, i+pos );
		}
		if( i > len ){
			len = i;
		}
		c = input_char();
		if( c >= 0x20 ){
			change_flag = CHANGE;
			if( mode == INSERT ){
				if( len + 1 >= max ){
					continue;
				}
				memmove( &str[i+1], &str[i], max - i - 1 );
				str[i] = (char)c;
				putchar( str[i++] );
				breaked = flush_after( str, i, ++len, pos, breaked );
				move_cursor( len+pos, i+pos );
			} else {
				str[i] = (char)c;
				putchar( str[i++] );
			}
		} else {
			switch( c ){
			case BS:
				if( mode == INSERT && i > 0 ){
					change_flag = CHANGE;
					memmove( &str[i-1], &str[i], max-i+1);
					move_cursor( i+pos, i - 1+pos );
					flush_after( str, --i, --len, pos, 2 );
					if( (len+1+pos) % columns == 0 ){
						putchar( ' ' );
						move_cursor( len+2+pos, i+pos );
					} else {
						move_cursor( len+1+pos, i+pos );
					}
				} else {
					if( i > 0 ){
						change_flag = CHANGE;
						move_cursor( i+pos, i - 1+pos );
						i--;
					}
				}
				break;
			case RET:
				move_cursor( i+pos, len+pos );
				putchar( '\n' );
				str[len] = '\0';
				prev = prevtmp;
				at = attmp;
				next = nexttmp;
				if( replace_dash( str ) ){
					len = strlen( str );
				}
				set_ring( str, strlen( str ) );
				change_flag = NOCHANGE;
				str[len] = '\n';
				str[len+1] = '\0';
				return;
				break;
			case LEFT:
				if( i > 0 ){
					move_cursor( i+pos, i - 1+pos );
					i--;
				}
				break;
			case RIGHT:
				if( i < max ){
					if( str[i] == 0 ){
						str[i] = ' ';
					}
					move_cursor( i+pos, i + 1+pos );
					i++;
				}
				break;
			case DLINE:
				change_flag = CHANGE;
				c_clear();
				memset( &str[i], '\0', max - i );
				len = i;
				break;
			case UP:
			case DOWN:
				strcpy( buff, str );
				get_ring( c, str, max );
				if( change_flag == CHANGE ){
					set_ring( buff, strlen( buff ) );
				}
				change_flag = NOCHANGE;
				move_cursor( i+pos, pos );
				c_clear();
				len = strlen( str );
				i = len;
				breaked = flush_after( str, 0, len, pos, breaked );
				break;
			case TAB:
				change_flag = CHANGE;
				if( mode == INSERT ){
					tabn = min( TABSIZE, max - len );
					memmove( &str[i+tabn], &str[i], max - i - tabn );
					memset( &str[i], ' ', tabn );
					len += tabn;
					breaked = flush_after( str, i, len, pos, breaked );
					i += tabn;
					move_cursor( j+pos, i+pos );
				} else {
					tabn = min( TABSIZE, max - i );
					memset( &str[i], ' ', tabn );
					i += tabn;
					for( j = 0; j < tabn; j++ ){
						putchar( ' ' );
					}
				}
				break;
			case DEL:
				change_flag = CHANGE;
				if( mode == INSERT ){	
					if( i >= len ){
						break;
					}
					memmove( &str[i], &str[i+1], max - i );
					flush_after( str, i, --len, pos, 2 );
					if( (len+1+pos) % columns == 0 ){
						putchar( ' ' );
						move_cursor( len+2+pos, i+pos );
					} else {
						move_cursor( len+1+pos, i+pos );
					}
				} else {
					str[i] = ' ';
					putchar( ' ' );
					move_cursor( i+1+pos, i+pos );
				}
				break;
			case INS:
				if( mode == INSERT ){
					mode = OVERW;
				} else {
					mode = INSERT;
				}
				break;
			case TOP:
				move_cursor( i+pos, pos );
				i = 0;
				break;
			case TAIL:
				for( j = len - 1; j >= 0; j-- ){
					if( str[j] != ' ' && str[j] != 0 ){
						break;
					}
				}
				j++;
				move_cursor( i+pos, j+pos );
				i = j;
				break;
			default:
				break;
			}
		}
	}
}
